# This source code is part of the proteusAI package and is distributed
# under the MIT License.

"""
A subpackage for sklearn_tools activity_prediction.
"""

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from proteusAI.visual_tools.plots import *  # noqa: F403
