# This source code is part of the proteusAI package and is distributed
# under the MIT License.

"""
A subpackage is concerned with the Model object of ProteusAI.
"""

__name__ = "proteusAI"
__author__ = "Jonathan Funk"

from .model import Model  # noqa: F401
